---------------------------------------------------------------------
--  PROCESSADOR R7
--
--  moraes  - 07/marco/2001     -     inicio do projeto
--  MODIFICACAO - 22/abril/01 - NOVO BANCO DE REGISTRADORES ==> ARRAY
--  ney		- 21/agosto/2001	- reviso p/ implementacao pipe
--  MODIFICACAO - 21/08/01 - Instruo NOP incorreta, corrigido
--
---------------------------------------------------------------------

---------------------------------------------------------------------
-- package com tipos basicos
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

package r7 is
  subtype reg16  is std_logic_vector(15 downto 0);
   
  -- indica o tipo de instrucao decodificada pela unidade de controle
  type instrucao is 
  (add, sub, e,   ou,  oux,  addi, subi, ldl,  ldh,  pop, push, ld,
  st, sl, sr, notB, dif, eql, sup, inf, ldsp, nop, rts,  halt, jmpr,
  jsrr, jmp, jsr,  jmpd, jsrd); 
  
  type microinstrucao is record
     mpc:   std_logic_vector(1 downto 0);  
	 -- controle do mux de entrada do registrador PC
     msp:   std_logic_vector(1 downto 0);  
	 -- mux do sp (stack-pointer)
     mdt:   std_logic;                     
	 -- mux para origem dos dados (DT) que vao para a memoria
     mad:   std_logic_vector(1 downto 0);  
	 -- mux para origem do endereco (AD) da memoria
     mreg:  std_logic;                     
	 -- controle do dado que entra no banco de registradores
     ms2:   std_logic;                     
	 -- origem do segundo registrador fonte (source2 - s2)
     ma:    std_logic;                     
	 -- mux para a selecao do operando A na entrada da ULA
     mb:    std_logic;                     
	 -- mux para a selecao do operando B na entrada da ULA
     wpc:   std_logic;                     
	 -- ENABLE para escrita em PC
     wsp:   std_logic;
     wir:   std_logic;
     wab:   std_logic;
     wula:  std_logic;
     wflag: std_logic;
     wreg:  std_logic;
     ce,rw: std_logic;                     
	 -- Chip enable e RW - Write ativo em 0
     ula:   instrucao;
  end record;
         
  -- registrador de 16 bits, com reset e habilitacao de escrita
  component registrador
       port( ck,rst,ce:in std_logic;
             D:in  reg16;
             Q:out reg16 );
  end component;
end r7;

---------------------------------------------------------------------
-- Registrador de uso geral      -   SENSIVEL A BORDA DE DESCIDA
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r7.all;

entity registrador is
      port( ck,rst,ce:in std_logic;
            D:in  reg16;
            Q:out reg16 );
end registrador;

architecture registrador of registrador is
   begin
      process (ck, rst)
        begin
            if rst = '1' then
                  Q <= (others => '0');
            elsif ck'event and ck = '0' then
                  if ce = '1' then  Q <= D; end if;
            end if;
      end process;
end registrador;

---------------------------------------------------------------------
-- Banco de registradores (R0..R15) - OS 16 REGISTRADORES SAO USADOS
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use ieee.STD_LOGIC_UNSIGNED.all;   
use work.r7.all;

entity bcregs is
      port(  ck, rst, wreg, rs2: in   std_logic;
             ir, inREG:          in   reg16;
             source1, source2:   out  reg16	);
end entity;

architecture bcregs of bcregs is   
   type banco is array (0 to 15) of reg16;
   signal reg : banco;
   signal wen:   reg16;
   signal destB: std_logic_vector(3 downto 0);
begin            
     
    r1:for i in 0 to 15 generate   
        wen(i) <= '1' when  ir(11 downto 8)=i and wreg='1'  else '0';
        rx: registrador PORT MAP(ck=>ck, rst=>rst, ce=>wen(i), 
		d=>inREG, q=>reg(i));               
    end generate r1;       
  
    -- selecao do source1
    source1 <=  reg( CONV_INTEGER(ir(7 downto 4)) );

    -- selecao do source2
    destB <= ir(3 downto 0) when rs2 = '0' else ir(11 downto 8);
    source2 <=   reg( CONV_INTEGER(destB) );
   
end bcregs;

---------------------------------------------------------------------
---------------------------------------------------------------------
--  Descricao estrutural do bloco de dados
---------------------------------------------------------------------
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;
use work.r7.all;
   
entity datapath is
      port( uins: in microinstrucao;
            ck, rst: in std_logic;
            instrucao, endereco : out reg16;
            data: inout reg16;
            flag: out std_logic );
end datapath;

architecture datapath of datapath is

    component bcregs
          port( ck, rst, wreg, rs2: in std_logic;
                ir, inREG:           in reg16;
                source1, source2:    out reg16 );
    end component;

    signal dtreg, dtpc, dtsp, s1, s2, outula, pc, sp, ir, rA, rB,
           rula, opA, opB, dataout, extended8, extended4: reg16;
    signal iflag, outcmp: std_logic;
begin

  --  conexao do registrador ir ao sinal de saida 'instrucao'
  instrucao <= ir;

  --
  --  registradores do bloco de dados: banco, pc, sp, ir, rA, rB,rula
  --
  REGS: bcregs port map( ck=>ck, rst=>rst, wreg=>uins.wreg, 
  rs2=>uins.ms2, ir=>ir, inREG=> dtreg, source1=>s1,
  source2=>s2);

  RPC:     registrador port map(ck=>ck, rst=>rst, ce=>uins.wpc,
           d=>dtpc, q=>pc );
  RSP:     registrador port map(ck=>ck, rst=>rst, ce=>uins.wsp,
           d=>dtsp, q=>sp );
  RIR:     registrador port map(ck=>ck, rst=>rst, ce=>uins.wir,
           d=>data, q=>ir );
  REG_A:   registrador port map(ck=>ck, rst=>rst, ce=>uins.wab,
           d=>s1, q=>rA );
  REG_B:   registrador port map(ck=>ck, rst=>rst, ce=>uins.wab,
           d=>s2, q=>rB );
  REG_ULA: registrador port map(ck=>ck, rst=>rst, ce=>uins.wula,
           d=>outula, q=>rula );
  
  --
  --  multiplexadores
  --
  dtpc <= rula when uins.mpc = "00" else 
  -- selecao  de operando para o registrador PC
  data when uins.mpc = "01" else
            pc+1;         -- por default o PC e incrementado;

  dtsp <= ra   when uins.msp = "00" else
  -- selecao  de operando para o registrador SP
            sp-1 when uins.msp = "10" else
            sp+1;

  endereco  <= rula  when uins.mad = "00" else
  -- selecao de quem endereca a RAM
               pc    when uins.mad = "01" else
               sp;

  opA <= ir     when uins.ma = '1' else ra;
  -- operando A para a ULA
  opB <= pc     when uins.mb = '1' else rb;
  -- operando B para a ULA

  dtreg <= data when uins.mreg = '1' else rula;
  -- selecao do conteudo que sera escrito no banco de registradores
  dataout <= pc when uins.mdt = '1'  else s2; 
  -- selecao de quem ira' escrever na memoria
  data <= dataout when uins.ce='1' and uins.rw='0' 
          else (others => 'Z');

  ---
  ---  COMPARADOR - utiliza-se o campo IR[11..8] para 
  ---  selecao de qual comparacao sera' efetuada
  ---
  outcmp <=  '1' when (ir(11 downto 8)=0 and rA/=rB ) or 
                      (ir(11 downto 8)=1 and rA=rB ) or
                      (ir(11 downto 8)=2 and rA>rB  ) or
		              (ir(11 downto 8)=3 and rA<rB ) else
             '0';  
  process (ck, rst)
     begin
            if rst = '1' then
                  iflag <= '0';
            elsif ck'event and ck = '0' then
				if uins.wflag = '1' then  iflag <=outcmp;  end if;
            end if;
  end process;
  flag <= iflag;
  
  --
  --  constante e extensao de sinal, para imediatos e saltos
  --
  extended8 <= "00000000" & opA(7 downto 0);
  extended4 <= opA(11) & opA(11) & opA(11) & opA(11)
               & opA(11 downto 0);

  ---
  ---  ULA - operacoes so' dependem da instrucao corrente 
  --- (decodificada no bloco de controle)
  ---
  outula <=  opA + opB                          when uins.ula=add  or uins.ula=ld or uins.ula=st or 
                                                     uins.ula=jmpr or uins.ula=jsrr   else
             opA - opB                          when uins.ula=sub  else
             opA or opB                         when uins.ula=ou   else
             opA and opB                        when uins.ula=e    else
             opA xor opB                        when uins.ula=oux  else
             opB + extended8                    when uins.ula=addi else
             opB - extended8                    when uins.ula=subi else
             opB(15 downto 8) & opA(7 downto 0) when uins.ula=ldl  else
             opA(7 downto 0)  & opB(7 downto 0) when uins.ula=ldh  else
             opB(14 downto 0) & iflag           when uins.ula=sl   else
             iflag & opB(15 downto 1)           when uins.ula=sr   else
             not opB(15 downto 0)               when uins.ula=notB else
             opB + extended4                    when uins.ula=jmpd or uins.ula=jsrd  else
             opA;                               -- DEFAULT: passaA

end datapath;

---------------------------------------------------------------------
---------------------------------------------------------------------
--  Descricao do bloco de controle
---------------------------------------------------------------------
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;
use work.r7.all;
entity controle is
      port (uins:    out microinstrucao;
            rst,ck: in std_logic;
            flag:   in std_logic;
            ir:     in reg16 );
end controle;

architecture controle of controle is

  type type_state  is (busca, rreg, HALT, alt_sp, ula, comp,
    erts, epop, eload, estore, wbck, altera_pc,
    epush, ejsr);   -- 14 estados
  signal EA, PE :  type_state;

  signal inst_la1, inst_la2, inst_jmp, inst_jsr:  std_logic;
  signal i : instrucao; 
begin

  -------------------------------------------------------------------
  -- BLOCO (1/3) - DECODIFICAO DAS INSTRUES
  -------------------------------------------------------------------
  i <= add   when ir(15 downto 12)=0 else
       sub   when ir(15 downto 12)=1 else
       e     when ir(15 downto 12)=2 else
       ou    when ir(15 downto 12)=3 else
       oux   when ir(15 downto 12)=4 else
       addi  when ir(15 downto 12)=5 else
       subi  when ir(15 downto 12)=6 else
       ldl   when ir(15 downto 12)=7 else
       ldh   when ir(15 downto 12)=8 else
       pop   when ir(15 downto 12)=9 and ir(7downto 4)=0 else
       push  when ir(15 downto 12)=9 and ir(7downto 4)=1 else
       ld    when ir(15 downto 12)=10 else
       st    when ir(15 downto 12)=11 else
       sl    when ir(15 downto 12)=12 and ir(7downto 4)=0 else
       sr    when ir(15 downto 12)=12 and ir(7downto 4)=1 else
       notB  when ir(15 downto 12)=12 and ir(7downto 4)=2 else
       dif   when ir(15 downto 12)=13 and ir(11 downto 8)=0  else 
       eql   when ir(15 downto 12)=13 and ir(11 downto 8)=1  else 
       sup   when ir(15 downto 12)=13 and ir(11 downto 8)=2  else 
       inf   when ir(15 downto 12)=13 and ir(11 downto 8)=3  else 
       ldsp  when ir(15 downto 12)=13 and ir(11 downto 8)=4  else 
       nop   when ir(15 downto 12)=13 and ir(11 downto 8)=5  else 
       rts   when ir(15 downto 12)=13 and ir(11 downto 8)=6  else 
       halt  when ir(15 downto 12)=13 and ir(11 downto 8)=7  else 
       jmpr  when ir(15 downto 12)=13 and ir(11 downto 8)=8  else 
       jsrr  when ir(15 downto 12)=13 and ir(11 downto 8)=9  else 
       jmp   when ir(15 downto 12)=13 and ir(11 downto 8)=10 else 
       jsr   when ir(15 downto 12)=13 and ir(11 downto 8)=11 else 
       jmpd  when ir(15 downto 12)=14 else
       jsrd  when ir(15 downto 12)=15 else
	   nop; -- Nunca deve ocorrer, mas fica aqui para dar estrutura.
 
  uins.ula <= i;       --- ******** operacao que a ula ira' executar   

  -- instrucoes logico_aritmeticas (la) do tipo 1:
  inst_la1 <= '1' when i=add or i=sub or i=e or i=ou or i=oux or i=sl
	   or i=sr or i=notB else '0';

  -- instrucoes logico_aritmeticas (la) do tipo 2 (rt esta no lado
  -- direito da expressao):
  inst_la2 <= '1' when i=addi or i=subi or i=ldl or i=ldh else
              '0'; 
       
  -- instrucoes de salto (jump)  
  -- ***** OBSERVAR QUE AQUI JA' AVALIAMOS O FLAG *****
  inst_jmp <= '1' when (i=jmpr or i=jmp or i=jmpd) and flag='1' else
              '0';
  
  -- instrucoes de salto a subrotin (jump)
  -- ***** OBSERVAR QUE AQUI JA' AVALIAMOS O FLAG *****
  inst_jsr <= '1' when (i=jsrr or i=jsr or i=jsrd) and flag='1' else
              '0';
       
  -------------------------------------------------------------------
  -- BLOCO (2/3) -  controle dos multiplexadores
  -------------------------------------------------------------------

  uins.mpc <= "10" when EA=busca else
              "01" when EA=erts  else
              "00";

  uins.msp <= "10" when inst_jsr='1' or i=push else -- pos-decremento
              "01" when i=pop   or i=rts  else    -- pre-incremento
              "00";

  uins.mad <= "10" when EA=epop or EA=erts or EA=ejsr or EA=epush else   -- subrotina SP endereca
              "01" when EA=busca else                                    -- na busca o PC endereca
              "00";                                                      -- por default: LD/ST

  uins.mdt <= inst_jsr;         -- escreve-se na memoria o PC quando houver chamada 'a subrotina
 
  -- escreve nos registradores o conteudo vindo da memoria
  uins.mreg <= '1' when i=ld or i=pop else '0';
       
  -- o segundo fonte (source2) recebe o endereco do reg. destino quando for 
  -- uma operacao log_arimetica do tipo 2, push ou operacao de escrita na memoria, 
  -- pois deve-se enviar o conteudo do reg. para a RAM
  uins.ms2 <= '1' when inst_la2='1' or i=push or EA=estore else '0';
 
  -- primeiro operando da ULA e' o IR quando for operacao log_arimetica do tipo 2 ou
  -- jump/jsr com deslocamento curto
  uins.ma <= '1' when inst_la2='1' or i=jmpd or i=jsrd else '0';   

  uins.mb <= inst_jmp  or inst_jsr;           -- em jumps/jsr o PC sera o segundo operando da ULA

  -------------------------------------------------------------------
  -- BLOCO (3/3) -  MAQUINA DE ESTADOS DE CONTROLE - gera os comandos
  -- de escrita e acesso  RAM
  -------------------------------------------------------------------    
  process(rst, ck)
   begin
      if rst='1' then 
            EA <= busca;
       elsif ck'event and ck='1' then
            EA <= PE;
      end if;
  end process;
  
  process(EA,i) 
  -- CALCULA O PROXIMO estado do circuito 
  -- depende do ESTADO ATUAL e da INSTRUCAO CORRENTE
   begin

    case EA is

     --
     -- primeiro ciclo de clock
     --
     when busca =>  uins.wpc<='1';  uins.wsp<='0';  uins.wir<='1';  
	                uins.wreg <='0'; uins.wab<='0';  uins.wula<='0';
	                uins.wflag<='0'; uins.ce<='1';   uins.rw<='1';

                    if i=halt then      
				    -- encontrou HALT => nao gera mais microinst.
                         PE <= HALT;
                     else
                         PE <= rreg;
                    end if;

     --
     -- segundo ciclo de clock
     --
     when HALT  =>    uins.wpc<='0';         uins.wir<='0';
                      uins.ce<='0';          uins.rw<='0';                      
                      PE <=HALT;    -- NO HALT TRANCA E NAO MAIS SAI

     when rreg => uins.wpc<='0';
                      uins.wir<='0';
                      uins.ce<='0';
                      uins.rw<='0';
                      uins.wab<='1';
                      if i=rts or i=pop or i=ldsp then
                         PE <= alt_sp;
                      elsif i=dif or i=eql or i=sup or i=inf then
                         PE <= comp;
                      else
                         PE <= ula;
                      end if;

     --
     -- terceiro ciclo de clock
     --
     when comp => uins.wab<='0';
                      uins.wflag<='1';
                      PE <= busca;   
					  -- comparacoes executam em 3 ciclos 
					  -- ** ATENCAO **

     when alt_sp =>  uins.wab<='0';
                     uins.wsp<='1';
                     if i=rts then
                         PE <= erts;
                      elsif i=pop then
                         PE <= epop;
                      else
                         PE <= busca;   
						 -- ldsp executa em 3 ciclos e NAO em 4 
						 -- ** ATENCAO **
                     end if;

     when ula => uins.wab<='0';
                     uins.wula<='1';
                     if i=ld                             then     PE <= eload;
                      elsif i=st                         then     PE <= estore;
                      elsif inst_la1='1' or inst_la2='1' then     PE <= wbck;
                      elsif inst_jmp='1'                 then     PE <= altera_pc;
                      elsif inst_jsr='1'                 then     PE <= ejsr;
                      elsif i=push                       then     PE <= epush;
                      else  PE <= busca;   
					  -- o nop/jumps executam em 3 ciclos ** ATENCAO **
                     end if;

     --
     -- quarto ciclo de clock, estados:  erts, epop, eload, 
	 -- estore, wbck, altera_pc, ejsr, epush
     -- VAI TODO MUNDO DE VOLTA PARA O FETCH
     --
     when erts =>      uins.wsp<='0';     uins.wpc<='1';         -- ESCREVE NO PC
                       uins.ce<='1';      uins.rw<='1';          -- LE DA RAM 
                       PE <= busca;

     when epop =>      uins.wsp<='0';     uins.wreg<='1';       -- ESCREVE NOS REGISTRADORES
                       uins.ce<='1';      uins.rw<='1';         -- LE DA RAM 
                       PE <= busca;

     when eload =>     uins.wula<='0';    uins.wreg<='1';       -- ESCREVE NOS REGISTRADORES
                       uins.ce<='1';      uins.rw<='1';         -- LE DA RAM
                       PE <= busca;

     when estore =>    uins.wula<='0';
                       uins.ce<='1';      uins.rw<='0';         -- ESCREVE NA RAM
                       PE <= busca;

     when wbck =>      uins.wula<='0';    uins.wreg<='1';     -- ESCREVE NOS REGISTRADORES
                       PE <= busca;

     when altera_pc => uins.wula<='0';    uins.wpc<='1';   -- ESCREVE NOS REGISTRADORES
                       PE <= busca;

     when ejsr =>      uins.wula<='0';    
                       uins.wpc<='1';     uins.wsp<='1';         -- ESCREVE PC / SP
                       uins.ce<='1';      uins.rw<='0';          -- ESCREVE NA RAM
                       PE <= busca;

     when epush =>     uins.wula<='0';    uins.wsp<='1';         -- ESCREVE PC
                       uins.ce<='1';      uins.rw<='0';          -- ESCREVE NA RAM
                       PE <= busca;
   end case;

  end process;

end controle;

---------------------------------------------------------------------
-- Uniao dos componentes da r7
---------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r7.all;

entity processador is
      port(ck,rst: in std_logic;
            data: inout reg16;
            address: out reg16;
            ce,rw: out std_logic );
end processador;

architecture processador of processador is

      component controle
            port( uins:   out microinstrucao;
                  ck,rst: in std_logic;
                  flag:   in std_logic;
                  ir:     in reg16 );
      end component;

      component datapath
            port( uins: in microinstrucao;
                  ck, rst: in std_logic;
                  instrucao, endereco : out reg16;
                  data: inout reg16;
                  flag: out std_logic );
      end component;

      signal flag: std_logic;
      signal uins: microinstrucao;
      signal ir: reg16;

      begin

      dp: datapath   port map(uins=>uins, ck=>ck, rst=>rst,
            instrucao=>ir,endereco=>address,data=>data, flag=>flag);
      ctrl: controle port map(uins=>uins, ck=>ck, rst=>rst,
            flag=>flag, ir=>ir);

      ce <= uins.ce;
      rw <= uins.rw;

end processador;